/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.datasource.ca;

import org.diirt.datasource.CompositeDataSource;
import org.diirt.datasource.PVManager;

import com.aquenos.epics.jackie.diirt.datasource.AbstractJackieDataSourceProvider;
import com.aquenos.epics.jackie.diirt.datasource.JackieDataSource;

/**
 * <p>
 * Provider for {@link JackieDataSource}s that uses the name <em>ca</em>.
 * </p>
 * 
 * <p>
 * This provider should only be on the class-path if the JCA-based data-source
 * provider is not present. As the JCA-based data-source provider uses the same
 * name, both would collide and it would be a matter of chance which of two
 * would be used by the {@link CompositeDataSource}. This conflict is only
 * resolved when running in an OSGi environment (see below) or when this
 * data-source provider (or the JCA-based data-source provider) is explicitly
 * registered with the {@link CompositeDataSource}. The
 * {@link #registerWith(CompositeDataSource)} method should be called for this
 * purpose.
 * </p>
 * 
 * <p>
 * When the JCA-based data-source provider is also on the class-path, the
 * {@code JackieDataSourceProvider} can be used instead.
 * </p>
 * 
 * <p>
 * When this class is used in an OSGi environment, this bundle's activator
 * explicitly adds this data-source provider to the {@link CompositeDataSource}
 * returned by {@link PVManager#getDefaultDataSource()} (only if it actually is
 * a {@link CompositeDataSource}, this means if it has not been replaced by a
 * different implementation). In this case, this data-source provider is going
 * to be used when a channel requests the "ca" data-source provider, even if
 * another provider can also be discovered on the class-path. In such a
 * situation, it is safe to have both this and the JCA-based data-source
 * provider on the class-path, but the JCA-based data-source provider is not
 * going to be available through means of the {@link CompositeDataSource}.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class CaDataSourceProvider extends AbstractJackieDataSourceProvider {

    /**
     * <p>
     * Explicitly registers this data-source provider with the specified
     * composite data-source. This is useful when both this data-source provider
     * and another data-source provider using the "ca" name are on the
     * class-path. In this case, this method should be called before using the
     * {@link CompositeDataSource} in order to ensure that the
     * {@code CompositeDataSource} consistently uses this provider.
     * </p>
     * 
     * <p>
     * When running in an OSGi environment, this bundle's activator
     * automatically calls this method passing the data-source returned by
     * {@link PVManager#getDefaultDataSource()} (as long as it actually is a
     * {@link CompositeDataSource}. This means that in an OSGi environment this
     * data-source provider will actually replace the standard "ca" data-source
     * provider as far as the {@link PVManager}'s default data-source is
     * concerned.
     * </p>
     * 
     * @param dataSource
     *            composite data-source with which this provider shall be
     *            registered.
     */
    public static void registerWith(CompositeDataSource dataSource) {
        dataSource.putDataSource(new CaDataSourceProvider());
    }

    @Override
    public String getName() {
        return "ca";
    }

}
