/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.vtype;

import java.text.NumberFormat;
import java.time.Instant;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.apache.commons.lang3.builder.ToStringBuilder;
import org.diirt.vtype.Display;

import com.aquenos.epics.jackie.common.value.ChannelAccessControlsValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessFloatingPointControlsValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessNumericControlsValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeValue;
import com.aquenos.epics.jackie.diirt.vtype.internal.NonLocalizedNumberFormats;

/**
 * Base class for all classes that represent numeric VTypes backed by Channel
 * Access values from EPICS Jackie. It implements the methods related to
 * meta-data (engineering units, limits, etc.).
 * 
 * @author Sebastian Marsching
 *
 * @param <ControlsValueType>
 *            type of the controls value. Must be derived from
 *            {@link ChannelAccessControlsValue}.
 * @param <TimeValueType>
 *            type of the time value. Must be derived from
 *            {@link ChannelAccessTimeValue}.
 */
public abstract class JackieDisplay<ControlsValueType extends ChannelAccessNumericControlsValue<?>, TimeValueType extends ChannelAccessTimeValue<?>>
        extends JackieAlarmAndTime<TimeValueType> implements Display {

    protected final ControlsValueType controlsValue;
    protected final boolean honorZeroPrecision;

    /**
     * <p>
     * Initializes this value with the specified backing value. The
     * {@code controlsValue} is used for the meta-data (engineering units,
     * limits, etc.) and the {@code timeValue} is used for the alarm status and
     * time-stamp.
     * </p>
     * 
     * <p>
     * If {@code disconnectTime} is not <code>null</code>, it is used as the
     * time-stamp and the alarm severity is set to {@code UNDEFINED}.
     * </p>
     * 
     * <p>
     * This constructor is package-private because the classes implementing the
     * VTypes form a closed hierarchy.
     * </p>
     * 
     * @param timeValue
     *            value used for alarm status and time-stamp.
     * @param disconnectTime
     *            time at which the channel was disconnected. Usually this
     *            should be <code>null</code>. If it is not <code>null</code>,
     *            the alarm severity is set to {@code UNDEFINED} and this
     *            parameter is used as the time-stamp (instead of using the
     *            time-stamp from the {@code timeValue}.
     * @param honorZeroPrecision
     *            defines how the value is formatted if the
     *            {@code controlsValue}'s precision is zero. If
     *            <code>true</code> a precision of zero means that no decimal
     *            digits shall be printed. If <code>false</code>, a precision of
     *            zero means that a default number of decimal digits shall be
     *            printed. This parameter is only relevant for the return value
     *            of the {@link #getFormat()} method.
     * @param scalar
     *            <code>true</code> if the child class represents a scalar,
     *            <code>false</code> if it represents an array.
     * @throws IllegalArgumentException
     *             if {@code scalar} is <code>true</code> and {@code timeValue}
     *             has more or less than one element.
     */
    JackieDisplay(ControlsValueType controlsValue, TimeValueType timeValue,
            Instant disconnectTime, boolean honorZeroPrecision,
            boolean scalar) {
        super(timeValue, disconnectTime, scalar);
        if (controlsValue == null) {
            throw new NullPointerException();
        }
        this.controlsValue = controlsValue;
        this.honorZeroPrecision = honorZeroPrecision;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null || !super.equals(obj)) {
            return false;
        }
        // The equals(...) implementation in the super-class already checks that
        // the types match, so we can cast without further tests.
        JackieDisplay<?, ?> other = (JackieDisplay<?, ?>) obj;
        return new EqualsBuilder()
                .append(this.controlsValue, other.controlsValue)
                .append(this.honorZeroPrecision, other.honorZeroPrecision)
                .build();
    }

    @Override
    public NumberFormat getFormat() {
        if (controlsValue instanceof ChannelAccessFloatingPointControlsValue) {
            ChannelAccessFloatingPointControlsValue<?> floatingPointValue = (ChannelAccessFloatingPointControlsValue<?>) controlsValue;
            short precision = floatingPointValue.getPrecision();
            if (precision < 0) {
                return NonLocalizedNumberFormats.getNumberFormatForPrecision(0);
            } else if (precision == 0) {
                // If the precision is zero it might be that it should actually
                // be zero or it might be unconfigured. We use the
                // honorZeroPrecision configuration setting in order to decide
                // how to format the value.
                if (honorZeroPrecision) {
                    return NonLocalizedNumberFormats
                            .getNumberFormatForPrecision(0);
                } else {
                    return NonLocalizedNumberFormats.getDefaultNumberFormat();
                }
            } else {
                return NonLocalizedNumberFormats
                        .getNumberFormatForPrecision(precision);
            }
        } else {
            // Integers always have a precision of zero.
            return NonLocalizedNumberFormats.getNumberFormatForPrecision(0);
        }
    }

    @Override
    public Double getLowerAlarmLimit() {
        return controlsValue.getGenericLowerAlarmLimit().doubleValue();
    }

    @Override
    public Double getLowerCtrlLimit() {
        return controlsValue.getGenericLowerControlLimit().doubleValue();
    }

    @Override
    public Double getLowerDisplayLimit() {
        return controlsValue.getGenericLowerDisplayLimit().doubleValue();
    }

    @Override
    public Double getLowerWarningLimit() {
        return controlsValue.getGenericLowerWarningLimit().doubleValue();
    }

    @Override
    public String getUnits() {
        return controlsValue.getUnits();
    }

    @Override
    public Double getUpperAlarmLimit() {
        return controlsValue.getGenericUpperAlarmLimit().doubleValue();
    }

    @Override
    public Double getUpperCtrlLimit() {
        return controlsValue.getGenericUpperControlLimit().doubleValue();
    }

    @Override
    public Double getUpperDisplayLimit() {
        return controlsValue.getGenericUpperDisplayLimit().doubleValue();
    }

    @Override
    public Double getUpperWarningLimit() {
        return controlsValue.getGenericUpperWarningLimit().doubleValue();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder(97, 31).appendSuper(super.hashCode())
                .append(controlsValue).append(honorZeroPrecision).build();
    }

    @Override
    public String toString() {
        return new ToStringBuilder(this).append("controlsValue", controlsValue)
                .append("disconnectTime")
                .append("honorZeroPrecision", honorZeroPrecision)
                .append("scalar", scalar).append("timeValue", timeValue)
                .build();
    }

}
